/*
 * Ctunnel - Cyptographic tunnel.
 * Copyright (C) 2008-2010 Jess Mahan <ctunnel@nardcore.org>
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <time.h>
#include <fcntl.h>
#include <signal.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <pthread.h>

#include "config.h"
#include "ctunnel.h"
#include "comp.h"
#include "crypto.h"
#include "net.h"
#include "opt.h"
#include "thread.h"

void crypto_shutdown(int sig)
{
    ctunnel_log(stdout, LOG_INFO, "Shutting Down\n");
#ifdef HAVE_OPENSSL
    EVP_cleanup();
#endif
    exit(0);
}

int main(int argc, char *argv[])
{
    int srv_sockfd = 0;
    int ret = 0;
#ifndef _WIN32
    struct sockaddr_in pin;
    socklen_t addrsize = 0;
    pid_t pid;
#endif
//    void *data = NULL;
    struct options opt;
    struct Network *net_cli = NULL, *net_srv = NULL;
    struct Ctunnel ctunnel;
    pthread_t tid;

    fprintf(stdout, "%s %d.%d %s\n", SOFTWARE_NAME, VERSION_MAJOR,
	    VERSION_MINOR, COPYRIGHT);

#ifndef _WIN32
    openlog("ctunnel", LOG_NDELAY | LOG_PID, LOG_DAEMON);
    ctunnel_log(stdout, LOG_INFO, "ctunnel %d.%d starting", VERSION_MAJOR,
		VERSION_MINOR);
#endif

    memset(opt.host, 0x00, sizeof(opt.host));
    opt = get_options(argc, argv);

#ifndef _WIN32
    signal(SIGCHLD, SIG_IGN);
    signal(SIGINT, crypto_shutdown);
    signal(SIGTERM, crypto_shutdown);
#endif

    if (opt.proto == 0) {
	srv_sockfd = tcp_listen(opt.port_listen);
	net_cli = calloc(1, sizeof(struct Network));
	net_srv = calloc(1, sizeof(struct Network));
    } else {
	net_srv = udp_bind(opt.port_listen);
	net_cli = udp_connect(opt.host, opt.port_forward);
    }

    if (opt.proto == 0) {
#ifdef HAVE_OPENSSL
	ctunnel_log(stdout, LOG_INFO,
		    "TCP [OpenSSL] Listening on %d using %s",
		    opt.port_listen, opt.key.ciphername);
#else
	ctunnel_log(stdout, LOG_INFO,
		    "TCP [libgcrypt] Listening on %d using %s %s",
		    opt.port_listen, opt.key.ciphername, opt.key.mode);
#endif
    } else {
#ifdef HAVE_OPENSSL
	ctunnel_log(stdout, LOG_INFO,
		    "UDP [OpenSSL] Listening on %d using %s",
		    opt.port_listen, opt.key.ciphername);
#else
	ctunnel_log(stdout, LOG_INFO,
		    "UDP [libgcrypt] Listening on %d using %s %s",
		    opt.port_listen, opt.key.ciphername, opt.mode);
#endif
    }

    if (opt.comp == 1)
	ctunnel_log(stdout, LOG_INFO, "zlib compression enabled [level %d]",
		    opt.comp_level);

    if (srv_sockfd < 0) {
	ctunnel_log(stderr, LOG_CRIT, "listen() %s\n", strerror(errno));
	exit(1);
    }
#ifndef _WIN32
    if (opt.daemon != 1) {
	ctunnel_log(stdout, LOG_INFO, "Going into background\n");
	pid = fork();
	if (pid == -1) {
	    ctunnel_log(stdout, LOG_CRIT, "fork() %s\n", strerror(errno));
	    exit(1);
	}
	if (pid > 0) {
	    exit(0);
	}
    }
#endif

    while (1) {
	if (opt.proto == 0) {
#ifdef _WIN32
	    net_srv->sockfd = accept(srv_sockfd, NULL, NULL);
#else
	    net_srv->sockfd = accept(srv_sockfd, (struct sockaddr *) &pin,
				     &addrsize);
#endif
	    net_cli->sockfd = tcp_connect(opt.host, opt.port_forward);
	}
	ctunnel.clisockfd = net_cli->sockfd;
	ctunnel.srvsockfd = net_srv->sockfd;
	ctunnel.opt = opt;
	ctunnel.srv_sockfd = srv_sockfd;

	if (opt.proto == 0) {
	    ret = pthread_create(&tid, NULL, (void *) ctunnel_mainloop,
				 (void *) &ctunnel);
	} else {
	    ctunnel.net_cli = net_cli;
	    ctunnel.net_srv = net_srv;
	    ctunnel_mainloop((void *) &ctunnel);
	}
    }
    return 0;
}
