/*
 * Ctunnel - Cyptographic tunnel.
 * Copyright (C) 2008-2010 Jess Mahan <ctunnel@nardcore.org>
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <time.h>
#include <fcntl.h>
#include <signal.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

#include <pthread.h>

#include "config.h"
#include "ctunnel.h"
#include "comp.h"
#include "crypto.h"
#include "net.h"

void *ctunnel_mainloop(void *arg)
{
    struct Ctunnel *ctunnel = (struct Ctunnel *) arg;
    struct options opt = (struct options) ctunnel->opt;
    struct Network *net_srv;
    struct Network *net_cli;
    struct Compress comp;
    struct Packet pkt_in = { NULL, 0 }, pkt_out = { NULL, 0 };
    unsigned char *data;
    fd_set rfd;
    int ret = 0;
#ifdef HAVE_OPENSSL
    crypto_ctx *ectx;
    crypto_ctx *dctx;
    struct Packet (*do_encrypt) (crypto_ctx * ctx, unsigned char *intext,
				 int size, unsigned char *key,
				 unsigned char *iv);
    do_encrypt = openssl_do_encrypt;
    struct Packet (*do_decrypt) (crypto_ctx * ctx, unsigned char *intext,
				 int size, unsigned char *key,
				 unsigned char *iv);
    do_decrypt = openssl_do_decrypt;
#else
    crypto_ctx ectx = NULL;
    crypto_ctx dctx = NULL;
    struct Packet (*do_encrypt) (crypto_ctx ctx, unsigned char *intext,
				 int size, unsigned char *key,
				 unsigned char *iv);
    do_encrypt = gcrypt_do_encrypt;
    struct Packet (*do_decrypt) (crypto_ctx ctx, unsigned char *intext,
				 int size, unsigned char *key,
				 unsigned char *iv);
    do_decrypt = gcrypt_do_decrypt;
#endif

    if (opt.proto == 0) {
	net_srv = calloc(1, sizeof(struct Network));
	net_cli = calloc(1, sizeof(struct Network));
	net_cli->sockfd = (int) ctunnel->clisockfd;
	net_srv->sockfd = (int) ctunnel->srvsockfd;
    } else {
	net_srv = (struct Network *) ctunnel->net_srv;
	net_cli = (struct Network *) ctunnel->net_cli;
    }

#ifdef HAVE_OPENSSL
    ectx = openssl_crypto_init(opt, 1);
    dctx = openssl_crypto_init(opt, 0);
#else
    ectx = gcrypt_crypto_init(opt);
    dctx = gcrypt_crypto_init(opt);
#endif


    if (opt.comp == 1)
	comp = z_compress_init(opt);

    data = calloc(1, sizeof(char *) * opt.packet_size);

    while (1) {
	if (opt.proto == 1) {
#ifdef HAVE_OPENSSL
	    EVP_CIPHER_CTX_cleanup(ectx);
	    EVP_CIPHER_CTX_cleanup(dctx);
	    free(ectx);
	    free(dctx);
	    ectx = openssl_crypto_init(opt, 1);
	    dctx = openssl_crypto_init(opt, 0);
#else
	    gcry_cipher_close(ectx);
	    gcry_cipher_close(dctx);
	    ectx = gcrypt_crypto_init(opt);
	    dctx = gcrypt_crypto_init(opt);
#endif
	}

	FD_ZERO(&rfd);
	FD_SET(net_srv->sockfd, &rfd);
	FD_SET(net_cli->sockfd, &rfd);
	ret = select(net_srv->sockfd + net_cli->sockfd + 1, &rfd,
		     NULL, NULL, NULL);

	if (FD_ISSET(net_srv->sockfd, &rfd)) {
	    ret = net_read(opt, net_srv, (char *) data, opt.packet_size);
	    if (ret <= 0) 
		break;
	    if (opt.server == 0) {
		if (opt.comp == 1) {
		    pkt_in = do_encrypt(ectx, data, ret,
			                opt.key.key, opt.key.iv);
		    pkt_out = z_compress(comp.zip, pkt_in.data,
					 pkt_in.len);
		} else {
		    pkt_out = do_encrypt(ectx, data, ret,
			                 opt.key.key, opt.key.iv);
		}
	    } else {
		if (opt.comp == 1) {
		    pkt_in = z_uncompress(comp.unzip, data, ret);
		    pkt_out = do_decrypt(dctx, pkt_in.data,
					 pkt_in.len, opt.key.key,
					 opt.key.iv);
		} else {
		    pkt_out = do_decrypt(dctx, data, ret,
			                 opt.key.key, opt.key.iv);
		}
	    }
	    ret = net_write(opt, net_cli, (char *) pkt_out.data, pkt_out.len);
	    if (ret <= 0)
		fprintf(stdout, "net_write %s:%d\n", __FILE__, __LINE__);

	    if (opt.comp == 1)
	        free(pkt_in.data);
	    free(pkt_out.data);

	    memset(data, 0x00, sizeof(char *) * opt.packet_size);
	}
	if (FD_ISSET(net_cli->sockfd, &rfd)) {
	    ret = net_read(opt, net_cli, (char *) data, opt.packet_size);
	    if (ret <= 0) 
		break;
	    if (opt.server == 0) {
		if (opt.comp == 1) {
		    pkt_in = z_uncompress(comp.unzip, data, ret);
		    pkt_out = do_decrypt(dctx, pkt_in.data, pkt_in.len,
					 opt.key.key, opt.key.iv);
		} else {
		    pkt_out = do_decrypt(dctx, data, ret,
					 opt.key.key, opt.key.iv);
		}
	    } else {
		if (opt.comp == 1) {
		    pkt_in = do_encrypt(ectx, data, ret,
					opt.key.key, opt.key.iv);
		    pkt_out = z_compress(comp.zip, pkt_in.data, pkt_in.len);
		} else {
		    pkt_out = do_encrypt(ectx, data, ret,
					 opt.key.key, opt.key.iv);
		}
	    }
	    ret = net_write(opt, net_srv, (char *) pkt_out.data, pkt_out.len);
	    if (ret <= 0)
		fprintf(stdout, "net_write %s:%d\n", __FILE__, __LINE__);

	    if (opt.comp == 1)
	        free(pkt_in.data);
	    free(pkt_out.data);

	    memset(data, 0x00, sizeof(char *) * opt.packet_size);
	}
	if (opt.proto == 1 && opt.comp == 1) {
	    deflateEnd(&comp.zip);
	    inflateEnd(&comp.unzip);
	    comp = z_compress_init(opt);
	}
	memset(data, 0x00, sizeof(char *) * opt.packet_size);
    }
    /* Connection closed */
    FD_CLR(net_cli->sockfd, &rfd);
    FD_CLR(net_srv->sockfd, &rfd);
    if (opt.proto == 0) {
	net_close(net_cli);
	net_close(net_srv);
#ifdef HAVE_OPENSSL
	EVP_CIPHER_CTX_cleanup(ectx);
	EVP_CIPHER_CTX_cleanup(dctx);
	EVP_cleanup();
#else
	gcry_cipher_close(ectx);
	gcry_cipher_close(dctx);
#endif
	if (opt.comp == 1) {
	    deflateEnd(&comp.zip);
	    inflateEnd(&comp.unzip);
	}

	free(net_cli);
	free(net_srv);
    }
    free(data);
    return NULL;
}
